(function() {
    let Strophe, $iq, $msg, $pres, $build, b64_sha1, _, Backbone, dayjs, _converse;
    let interestingServers = new Set();
    let subscribedServers = new Set();

    converse.plugins.add('converse-rai', {
        'dependencies': [],

        'initialize': function() {
            _converse = this._converse;

            Strophe = converse.env.Strophe;
            $iq = converse.env.$iq;
            $msg = converse.env.$msg;
            $pres = converse.env.$pres;
            $build = converse.env.$build;
            b64_sha1 = converse.env.b64_sha1;
            _ = converse.env._;
            Backbone = converse.env.Backbone;
            dayjs = converse.env.dayjs;

            _converse.api.settings.extend({
                rai_notification: true,
                rai_notification_label: 'Room Activity Indicator',
            });

            _converse.api.listen.on('connected', function() {
                setupRoomActivityIndicators();
            });

            _converse.api.listen.on('chatRoomViewInitialized', function(view) {
                const jid = view.model.get('jid');

                if (view.model.get('num_unread') > 0 || view.model.get('num_unread_general') > 0) {
                    emitNotification(jid);
                }
            });

            _converse.api.listen.on('raiRoomsUpdated', function(rooms) {
                interestingServers = new Set(rooms.filter(room => room).map(Strophe.getDomainFromJid));
                if (_converse.api.connection.connected()) {
                    updateSubscriptions();
                }
            });

            _converse.api.listen.on('chatBoxScrolledDown', function(view) {
                const jid = view.chatbox.get('jid');
                const id_attr = 'stanza_id ' + jid;
                const messages = view.chatbox.messages;

                if (!getUnreadStatus(jid)) {
                    return;
                }

                for (let i = messages.length - 1; i >= 0; i--) {
                    const message = messages.at(i);

                    if (message.has(id_attr)) {
                        let id = message.get(id_attr);
                        if (id != sessionStorage.getItem('rai_displayed.' + jid)) {
                            sessionStorage.setItem('rai_displayed.' + jid, id);
                            setUnreadStatus(jid, false);
                            setTimeout(() => sendMarker(jid, id, 'displayed'), 0);
                        }
                        break;
                    }
                }
            });

            _converse.api.listen.on('chatBoxInsertedIntoDOM', function(view) {
                const jid = view.model.get('jid');

                if (view.model.get('num_unread') > 0) {
                    emitNotification(jid);
                }

            });

            _converse.api.listen.on('message', function(data) {
                var chatbox = data.chatbox;
                var history = data.attrs.is_archived;
                var sender = data.attrs.sender;
                var body = data.attrs.body;

                if (!history && body && chatbox && sender !== 'me') {
                    const alert = chatbox.get('num_unread') > 0;
                    const notify = chatbox.get('num_unread_general') > 0;

                    if (alert || notify) {
                        emitNotification(chatbox.get('jid'), alert);
                    }
                }
            });
        },
    });


    function subscribeServer(server_name) {
        const id = Math.random().toString(36).substr(2, 9);
        _converse.connection.send(converse.env.$pres({
            to: server_name,
            id: id,
        }).c('rai', {
            'xmlns': 'xmpp:prosody.im/protocol/rai',
        }));
    }

    function unsubscribeServer(server_name) {
        _converse.connection.send(converse.env.$pres({
            to: server_name, type: 'unavailable',
        }).c('rai', {
            'xmlns': 'xmpp:prosody.im/protocol/rai',
        }));
    }

    function updateSubscriptions() {
        var new_servers = new Set([...interestingServers].filter(server => !subscribedServers.has(server)));
        var obsolete_servers = new Set([...subscribedServers].filter(server => !interestingServers.has(server)));
        for (let server of obsolete_servers) {
            unsubscribeServer(server);
        }
        for (let server of new_servers) {
            subscribeServer(server);
        }
    }

    function setupRoomActivityIndicators() {
        updateSubscriptions();
        // If we already have unread notifications stored for this session, emit them now
        for (var i = 0; i < sessionStorage.length; i++) {
            if (sessionStorage.key(i).indexOf('rai_notify.') == 0) {
                const jid = sessionStorage.key(i).substring(11);
                emitNotification(jid);
            }
        }

        // Listen for incoming RAI from the server
        _converse.connection.addHandler(function(message) {
            const from_jid = message.attributes.from?.nodevalue;
            const room_jid = from_jid?.split('/')[0];
            const room = '';
            let ignore = false;
            for (let i = 0; i < _converse.chatboxes.models.length; i++) {
                if (_converse.chatboxes.models[i].id === room_jid) {
                    room = _converse.chatboxes.models[i].id;
                    break;
                }
            }
            if (room && from_jid && room_jid) {
                if (from_jid === room_jid + '/' + room.get('nick')) {
                    ignore = true;
                }
            }
            if (message && !ignore)
                message.querySelectorAll('activity').forEach(function(activity) {
                    if (activity && activity.namespaceURI == 'xmpp:prosody.im/protocol/rai') {
                        const jid = activity.textContent;
                        setUnreadStatus(jid, true);
                        emitNotification(jid);
                    }
                });
            return true;
        }, null, 'message');
    }

    function setUnreadStatus(jid, flag) {
        if (flag) {
            sessionStorage.setItem('rai_notify.' + jid, 'true');
        } else {
            sessionStorage.removeItem('rai_notify.' + jid);
        }
    }

    function getUnreadStatus(jid) {
        return sessionStorage.getItem('rai_notify.' + jid) == 'true';
    }

    function emitNotification(jid, alert) {
        _converse.api.trigger('chatRoomActivityIndicators', jid);
    }

    function sendMarker(to_jid, id, type) {
        const stanza = converse.env.$msg({
            'from': _converse.connection.jid,
            'id': Math.random().toString(36).substr(2, 9),
            'to': to_jid,
            'type': 'groupchat',
        }).c(type, {
            'xmlns': converse.env.Strophe.NS.MARKERS,
            'id': id,
        });

        _converse.api.send(stanza);
    }
})();
